from chia import instrumentation
from chia.components.datasets import DatasetFactory
from chia.components.datasets.dataset import Dataset
from chia import knowledge

import unittest
import functools
import networkx as nx
from tqdm import tqdm

import argparse


def validate(config: dict, deep: bool):
    observer = instrumentation.ObserverFactory.create(config={"name": "stream"})
    obs = instrumentation.NamedObservable("validate_dataset")
    obs.register(observer)

    print("Instantiating dataset")
    dataset: Dataset = DatasetFactory.create(config=config, observers=[observer])

    setups = dataset.setups()
    print(f"Testing {len(setups)}")
    for i, setup in enumerate(setups):
        print(f"Loading setup {i + 1} of {len(setups)}")
        validate_setup(dataset, setup, observers=[observer], deep=deep)


def validate_setup(dataset: Dataset, setup, observers, deep):
    dataset.setup(**setup)

    tc = unittest.TestCase()

    # Data
    train_pool_count = dataset.train_pool_count()
    test_pool_count = dataset.test_pool_count()
    tc.assertGreater(train_pool_count, 0, "There are train pools")
    tc.assertGreater(test_pool_count, 0, "There are test pools")
    tc.assertIsNotNone(dataset.namespace())

    train_sample_uids = [set()] * train_pool_count
    test_sample_uids = [set()] * test_pool_count

    # Examine pools individually
    all_concept_uids = set()
    for train_pool_index in range(train_pool_count):
        print(f"Testing train pool {train_pool_index + 1} of {train_pool_count}")
        train_pool = dataset.train_pool(train_pool_index, "label_gt")
        tc.assertGreater(len(train_pool), 0, "Train pool is not empty")

        for sample in tqdm(train_pool):
            validate_sample(tc, dataset, sample, deep)
            all_concept_uids |= {sample.get_resource("label_gt")}
            train_sample_uids[train_pool_index] |= {sample.get_resource("uid")}

    for test_pool_index in range(test_pool_count):
        print(f"Testing test pool {test_pool_index + 1} of {test_pool_count}")
        test_pool = dataset.test_pool(test_pool_index, "label_gt")
        tc.assertGreater(len(test_pool), 0, "Test pool is not empty")

        for sample in tqdm(test_pool):
            validate_sample(tc, dataset, sample, deep)
            all_concept_uids |= {sample.get_resource("label_gt")}
            test_sample_uids[test_pool_index] |= {sample.get_resource("uid")}

    # Train / test intersection
    all_train_uids = functools.reduce(lambda x, y: x | y, train_sample_uids, set())
    all_test_uids = functools.reduce(lambda x, y: x | y, test_sample_uids, set())

    train_test_intersection = all_train_uids.intersection(all_test_uids)

    tc.assertTrue(
        len(train_test_intersection) == 0, "Train and test set have no intersection"
    )

    # Relation
    tc.assertIsNotNone(
        dataset.get_hyponymy_relation_source(), "Dataset provides a hypernymy relation"
    )

    # Compare all_concept_uids to prediction targets
    prediction_target_uids = set(dataset.prediction_targets())
    for prediction_target_uid in prediction_target_uids:
        tc.assertIn(
            prediction_target_uid,
            all_concept_uids,
            "Dataset reports prediction targets correctly",
        )

    # Try constructing a knowledge base
    kb = knowledge.KnowledgeBaseFactory.create(dict(), observers=observers)
    kb.add_prediction_targets(dataset.prediction_targets())
    relation_sources = [dataset.get_hyponymy_relation_source()]
    relation_sources += [knowledge.WordNetAccess()]

    kb.add_hyponymy_relation(relation_sources)

    rgraph: nx.DiGraph = kb.get_hyponymy_relation_rgraph()
    possible_roots = [node for node in rgraph.nodes if rgraph.in_degree(node) == 0]
    tc.assertEqual(1, len(possible_roots), "There should be only one root.")
    root = next(nx.topological_sort(rgraph))

    for uid in all_concept_uids:
        # The root is fine
        if uid == root:
            continue

        tc.assertTrue(
            nx.has_path(rgraph, root, uid),
            f"The UID {uid} should be connected to the root {root}.",
        )


def validate_sample(tc, dataset, sample, deep):
    tc.assertTrue(
        str(sample.get_resource("uid")).startswith(f"{dataset.namespace()}::"),
        "Sample UID has correct namespace",
    )
    tc.assertTrue(
        str(sample.get_resource("label_gt")).startswith(f"{dataset.namespace()}::"),
        "Sample label has correct namespace",
    )
    if deep:
        # Attemt to load input image
        sample.get_resource("input_img_np")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    gp = parser.add_mutually_exclusive_group(required=True)
    gp.add_argument("--name", type=str)
    gp.add_argument("--json", type=str)
    parser.add_argument("--deep", action="store_true")
    args = parser.parse_args()

    if args.name is not None:
        validate({"name": args.name}, args.deep)

    elif args.json is not None:
        import config as pcfg

        configs = [pcfg.config_from_json(args.json, read_from_file=True)]
        config = pcfg.ConfigurationSet(*configs)

        validate(config, args.deep)



import argparse
import gzip
import json
import typing
import logging
from pathlib import Path
from datetime import datetime

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure
from PyQt5 import QtWidgets, uic

UI_FILENAME = Path(__file__).parent / "ui" / "json_result_viewer.ui"


class JSONResultViewerUI(QtWidgets.QWidget):
    def __init__(self, json_path):
        super().__init__()
        uic.loadUi(UI_FILENAME, self)

        self.json_path = json_path
        self.json_obj = None
        self.load_json()

    def load_json(self):
        if self.json_path.endswith(".json"):
            with open(self.json_path) as file_obj:
                self.json_obj = json.load(file_obj)
        elif self.json_path.endswith(".json.gz"):
            with gzip.open(self.json_path, "rt") as file_obj:
                self.json_obj = json.load(file_obj)

        # Assertions

        # Raw code
        self.setup_tab_raw_json()

        # Metadata
        self.setup_tab_metadata()

        # Metrics
        self.setup_tab_metrics()

        # Results
        self.setup_tab_result()

        # Logs
        self.setup_tab_log()

    def setup_tab_result(self):
        # Read combinations of metrics and senders
        message_type_dict = dict()
        for message_obj in self.json_obj["ResultMessage"]:
            sender_name = message_obj["sender"]
            for result_key, result_value in message_obj["result_dict"].items():
                key = (sender_name, result_key)
                if not (
                    isinstance(result_value, int) or isinstance(result_value, float)
                ):
                    # We don't want to display such items at all
                    continue
                if key in message_type_dict.keys():
                    message_type_dict[key] += 1
                else:
                    message_type_dict[key] = 1

        self.table_resulttypes.setRowCount(len(message_type_dict.keys()))
        for i, (key, value) in enumerate(message_type_dict.items()):
            self.table_resulttypes.setItem(i, 0, QtWidgets.QTableWidgetItem(key[0]))
            self.table_resulttypes.setItem(i, 1, QtWidgets.QTableWidgetItem(key[1]))
            self.table_resulttypes.setItem(i, 2, QtWidgets.QTableWidgetItem(str(value)))
        self.table_resulttypes.setColumnWidth(0, 150)
        self.table_resulttypes.setColumnWidth(1, 150)
        self.table_resulttypes.setColumnWidth(2, 50)

        self.table_resulttypes.itemSelectionChanged.connect(self.update_results)
        self.radio_result_step.clicked.connect(self.update_results)
        self.radio_result_timestamp.clicked.connect(self.update_results)

    def update_results(self):
        selection_ranges: typing.List[
            QtWidgets.QTableWidgetSelectionRange
        ] = self.table_resulttypes.selectedRanges()

        result_tuples = []
        for selection_range in selection_ranges:
            for row in range(selection_range.topRow(), selection_range.bottomRow() + 1):
                sender = self.table_resulttypes.item(row, 0).text()
                result = self.table_resulttypes.item(row, 1).text()
                xaxis = "step" if self.radio_result_step.isChecked() else "timestamp"

                result_tuples += [(sender, result, xaxis)]

        self.plot_results(result_tuples)
        self.display_raw_results(result_tuples)

    def plot_results(self, result_tuples):
        # Clear the layout
        layout = self.groupbox_result_plot.layout()
        child = layout.takeAt(0)
        if child is not None:
            layout.removeItem(child)
            del child

        # Plot to a new widget
        figure = Figure()
        axes = figure.add_subplot(111)
        layout.addWidget(FigureCanvasQTAgg(figure))

        # Plots
        for result_tuple in result_tuples:
            sender, result, xaxis = result_tuple
            # Get the data
            Xs, Ys = self.extract_results(result, sender, xaxis)

            label = f"{sender}.{result}"
            axes.plot(Xs, Ys, label=label)

        axes.legend()

    def display_raw_results(self, result_tuples):
        if len(result_tuples) != 1:
            self.table_resultraw.clear()
            self.table_resultraw.setColumnCount(0)
            self.table_resultraw.setRowCount(0)
        else:
            result_tuple = result_tuples[0]
            sender, result, xaxis = result_tuple
            # Get the data
            Xs, Ys = self.extract_results(result, sender, xaxis)

            label = f"{sender}.{result}"

            self.table_resultraw.setColumnCount(2)
            self.table_resultraw.setRowCount(len(Xs))
            self.table_resultraw.setHorizontalHeaderItem(
                0, QtWidgets.QTableWidgetItem(xaxis)
            )
            self.table_resultraw.setHorizontalHeaderItem(
                1, QtWidgets.QTableWidgetItem(label)
            )

            for i, (x, y) in enumerate(zip(Xs, Ys)):
                self.table_resultraw.setItem(i, 0, QtWidgets.QTableWidgetItem(str(x)))
                self.table_resultraw.setItem(i, 1, QtWidgets.QTableWidgetItem(str(y)))

    def extract_results(self, result, sender, xaxis):
        valid_messages = [
            message_obj
            for message_obj in self.json_obj["ResultMessage"]
            if message_obj["sender"] == sender
            and result in message_obj["result_dict"].keys()
        ]
        Xs = []
        Ys = []
        for valid_message in valid_messages:
            Ys += [valid_message["result_dict"][result]]
            Xs += [valid_message[xaxis]]
        return Xs, Ys

    def setup_tab_metrics(self):
        # Read combinations of metrics and senders
        message_type_dict = dict()
        for message_obj in self.json_obj["MetricMessage"]:
            if not (
                isinstance(message_obj["value"], int)
                or isinstance(message_obj["value"], float)
            ):
                # Filter out undisplayable things
                continue
            key = (message_obj["sender"], message_obj["metric"])
            if key in message_type_dict.keys():
                message_type_dict[key] += 1
            else:
                message_type_dict[key] = 1
        self.table_metrictypes.setRowCount(len(message_type_dict.keys()))
        for i, (key, value) in enumerate(message_type_dict.items()):
            self.table_metrictypes.setItem(i, 0, QtWidgets.QTableWidgetItem(key[0]))
            self.table_metrictypes.setItem(i, 1, QtWidgets.QTableWidgetItem(key[1]))
            self.table_metrictypes.setItem(i, 2, QtWidgets.QTableWidgetItem(str(value)))
        self.table_metrictypes.setColumnWidth(0, 150)
        self.table_metrictypes.setColumnWidth(1, 150)
        self.table_metrictypes.setColumnWidth(2, 50)

        self.table_metrictypes.itemSelectionChanged.connect(self.update_metrics)
        self.radio_metric_step.clicked.connect(self.update_metrics)
        self.radio_metric_timestamp.clicked.connect(self.update_metrics)

    def update_metrics(self):
        selection_ranges: typing.List[
            QtWidgets.QTableWidgetSelectionRange
        ] = self.table_metrictypes.selectedRanges()

        metric_tuples = []
        for selection_range in selection_ranges:
            for row in range(selection_range.topRow(), selection_range.bottomRow() + 1):
                sender = self.table_metrictypes.item(row, 0).text()
                metric = self.table_metrictypes.item(row, 1).text()
                xaxis = "step" if self.radio_metric_step.isChecked() else "timestamp"

                metric_tuples += [(sender, metric, xaxis)]

        self.plot_metrics(metric_tuples)
        self.display_raw_metrics(metric_tuples)

    def plot_metrics(self, metric_tuples):
        # Clear the layout
        layout = self.groupbox_metric_plot.layout()
        child = layout.takeAt(0)
        if child is not None:
            layout.removeItem(child)
            del child

        # Plot to a new widget
        figure = Figure()
        axes = figure.add_subplot(111)
        layout.addWidget(FigureCanvasQTAgg(figure))

        # Plots
        for metric_tuple in metric_tuples:
            sender, metric, xaxis = metric_tuple
            # Get the data
            Xs, Ys = self.extract_metrics(metric, sender, xaxis)

            label = f"{sender}.{metric}"
            axes.plot(Xs, Ys, label=label)

        axes.legend()

    def display_raw_metrics(self, metric_tuples):
        if len(metric_tuples) != 1:
            self.table_metricraw.clear()
            self.table_metricraw.setColumnCount(0)
            self.table_metricraw.setRowCount(0)
        else:
            metric_tuple = metric_tuples[0]
            sender, metric, xaxis = metric_tuple
            # Get the data
            Xs, Ys = self.extract_metrics(metric, sender, xaxis)

            label = f"{sender}.{metric}"

            self.table_metricraw.setColumnCount(2)
            self.table_metricraw.setRowCount(len(Xs))
            self.table_metricraw.setHorizontalHeaderItem(
                0, QtWidgets.QTableWidgetItem(xaxis)
            )
            self.table_metricraw.setHorizontalHeaderItem(
                1, QtWidgets.QTableWidgetItem(label)
            )

            for i, (x, y) in enumerate(zip(Xs, Ys)):
                self.table_metricraw.setItem(i, 0, QtWidgets.QTableWidgetItem(str(x)))
                self.table_metricraw.setItem(i, 1, QtWidgets.QTableWidgetItem(str(y)))

    def extract_metrics(self, metric, sender, xaxis):
        valid_messages = [
            message_obj
            for message_obj in self.json_obj["MetricMessage"]
            if message_obj["sender"] == sender and message_obj["metric"] == metric
        ]
        Xs = []
        Ys = []
        for valid_message in valid_messages:
            Ys += [valid_message["value"]]
            Xs += [valid_message[xaxis]]
        return Xs, Ys

    def setup_tab_metadata(self):
        self.line_experimentname.setText(self.json_obj["metadata"]["name"])
        self.line_runid.setText(self.json_obj["metadata"]["run_id"])
        self.label_chiaversion.setText(
            f'Using CHIA {self.json_obj["environment"]["chia_version"]}'
        )

        # Environment Variables
        environment_dict: dict = dict()
        for key in self.json_obj["environment"]["environment_variables"].keys():
            value = self.json_obj["environment"]["environment_variables"][key]
            if (
                isinstance(value, int)
                or isinstance(value, float)
                or isinstance(value, str)
                or isinstance(value, bool)
                or value is None
            ):
                environment_dict[key] = value
        self.table_environment.setColumnWidth(0, 300)
        self.table_environment.setRowCount(len(environment_dict.keys()))
        for i, (key, value) in enumerate(environment_dict.items()):
            self.table_environment.setItem(i, 0, QtWidgets.QTableWidgetItem(key))
            self.table_environment.setItem(i, 1, QtWidgets.QTableWidgetItem(value))

        # Configuration
        self.table_configuration.setRowCount(len(self.json_obj["ConfigMessage"]))
        self.table_configuration.setColumnWidth(0, 60)
        self.table_configuration.setColumnWidth(1, 300)
        for i, config_message in enumerate(self.json_obj["ConfigMessage"]):
            self.table_configuration.setItem(
                i, 0, QtWidgets.QTableWidgetItem(str(config_message["source"]))
            )
            self.table_configuration.setItem(
                i, 1, QtWidgets.QTableWidgetItem(str(config_message["field"]))
            )
            self.table_configuration.setItem(
                i, 2, QtWidgets.QTableWidgetItem(str(config_message["value"]))
            )

    def setup_tab_log(self):
        relevant_log_messages = self.json_obj["LogMessage"]

        self.table_log.setRowCount(len(relevant_log_messages))
        self.table_log.setColumnWidth(0, 180)
        self.table_log.setColumnWidth(1, 100)
        self.table_log.setColumnWidth(2, 280)

        for i, log_message in enumerate(relevant_log_messages):
            date_object = datetime.fromtimestamp(log_message["timestamp"])
            self.table_log.setItem(
                i,
                0,
                QtWidgets.QTableWidgetItem(
                    str(date_object.strftime("%Y-%m-%d %H:%M:%S"))
                ),
            )
            self.table_log.setItem(
                i,
                1,
                QtWidgets.QTableWidgetItem(
                    str(logging.getLevelName(log_message["level"]))
                ),
            )
            self.table_log.setItem(
                i, 2, QtWidgets.QTableWidgetItem(str(log_message["sender"]))
            )
            self.table_log.setItem(
                i, 3, QtWidgets.QTableWidgetItem(str(log_message["message"]))
            )

    def setup_tab_raw_json(self):
        reencoded = json.dumps(self.json_obj, indent=2)
        self.raw_json_text.setPlainText(reencoded)


def main(json_path):
    app = QtWidgets.QApplication([])
    window = JSONResultViewerUI(json_path)
    window.show()

    app.exec_()


if __name__ == "__main__":
    parser = argparse.ArgumentParser("chia.tools.json_result_viewer")
    parser.add_argument("json_path", type=str, help="Input file path")
    args = parser.parse_args()
    main(args.json_path)



import argparse
import gzip
import json
import pathlib
import typing

import pandas as pd


def aggregrate_metrics(
    metrics_per_step: typing.Dict[int, dict], max_accumulated_message_count: int
) -> typing.Dict[int, dict]:
    ret_dict = dict()
    accumulator = dict()
    count = dict()

    accumulated_messages = 0

    input_steps = list(sorted(metrics_per_step.keys()))
    for step in input_steps:
        step_input_dict = metrics_per_step[step]
        if step < 0:
            # Don't accumulate step -1, it containes different data
            ret_dict[step] = step_input_dict

        for key, value in step_input_dict.items():
            if key not in accumulator.keys():
                accumulator[key] = 0.0
                count[key] = 0
            accumulator[key] += value
            count[key] += 1

        accumulated_messages += 1

        # Write out aggregate every 100 messages and always for the last step
        if (
            accumulated_messages >= max_accumulated_message_count
            or step == input_steps[-1]
        ):
            step_aggregate_dict = dict()
            for key, acc_value in accumulator.items():
                assert 0 < count[key] <= accumulated_messages
                step_aggregate_dict[key] = acc_value / float(count[key])

            ret_dict[step] = step_aggregate_dict
            accumulated_messages = 0
            count = dict()
            accumulator = dict()

    output_steps = list(sorted(ret_dict.keys()))

    print(f"Accumulated {len(input_steps)} to {len(output_steps)} messages.")
    return ret_dict


def process_json(
    obj: dict, max_accumulated_message_count: int
) -> typing.Tuple[typing.List[dict], typing.List[dict]]:
    assert isinstance(obj, dict), "Only dictionary types are supported"

    # Get config dict
    config_dict: dict = dict()
    for config_obj in obj["ConfigMessage"]:
        value = config_obj["value"]
        if not (
            isinstance(value, int)
            or isinstance(value, float)
            or isinstance(value, str)
            or isinstance(value, bool)
            or value is None
        ):
            continue
        config_dict[config_obj["field"]] = value

    assert config_dict["RunnerFactory.name"] == "epoch"

    # Check shutdown message
    assert obj["ShutdownMessage"][0]["successful"]

    # Get metadata
    metadata_dict: dict = dict()
    for key in obj["metadata"].keys():
        value = obj["metadata"][key]
        if (
            isinstance(value, int)
            or isinstance(value, float)
            or isinstance(value, str)
            or isinstance(value, bool)
            or value is None
        ):
            metadata_dict[key] = value

    results_combined_with_metadata = list()

    for result_obj in obj["ResultMessage"]:
        single_result_dict: dict = dict()
        for key in result_obj["result_dict"].keys():
            value = result_obj["result_dict"][key]
            if (
                isinstance(value, int)
                or isinstance(value, float)
                or isinstance(value, str)
                or isinstance(value, bool)
                or value is None
            ):
                single_result_dict[key] = value
            elif isinstance(value, list):
                for i, element in enumerate(value):
                    single_result_dict[f"{key}_{i}"] = element
            else:
                continue

        single_result_dict["step"] = result_obj["step"]
        single_result_dict["sender"] = result_obj["sender"]
        single_result_dict["timestamp"] = result_obj["timestamp"]

        combined_dict = dict()
        combined_dict.update({f"conf_{k}": v for k, v in config_dict.items()})
        combined_dict.update({f"sres_{k}": v for k, v in single_result_dict.items()})
        combined_dict.update({f"meta_{k}": v for k, v in metadata_dict.items()})
        results_combined_with_metadata += [combined_dict]

    metric_per_step = dict()
    for result_obj in obj["MetricMessage"]:
        step = result_obj["step"]
        if step not in metric_per_step.keys():
            metric_per_step[step] = dict()

        metric_per_step[step][result_obj["metric"]] = result_obj["value"]
        metric_per_step[step]["step"] = step

    metric_per_step = aggregrate_metrics(metric_per_step, max_accumulated_message_count)

    metrics_combined_with_metadata = list()
    for step, metric_dict in metric_per_step.items():
        combined_dict = dict()
        combined_dict.update({f"conf_{k}": v for k, v in config_dict.items()})
        combined_dict.update({f"metr_{k}": v for k, v in metric_dict.items()})
        combined_dict.update({f"meta_{k}": v for k, v in metadata_dict.items()})

        metrics_combined_with_metadata += [combined_dict]

    return results_combined_with_metadata, metrics_combined_with_metadata


def process_json_path(
    path: str, max_accumulated_message_count: int
) -> typing.Tuple[typing.List[dict], typing.List[dict]]:
    if path.endswith(".json"):
        with open(path) as file_obj:
            return process_json(json.load(file_obj), max_accumulated_message_count)
    elif path.endswith(".json.gz"):
        with gzip.open(path, "rt") as file_obj:
            return process_json(json.load(file_obj), max_accumulated_message_count)
    else:
        raise ValueError(
            f"Filename ending of {path} is not supported (only .json and .json.gz are)."
        )


def main(
    json_paths: typing.List[str], output_path: str, max_accumulated_message_count: int
):
    # Check if output is available
    if pathlib.Path(output_path).exists():
        raise ValueError(f"File {output_path} already exists! Please delete it first.")

    # Read
    result_data = []
    metric_data = []
    for path in json_paths:
        print(f"Processing {path}...", end="")
        path_results, path_metrics = process_json_path(
            path, max_accumulated_message_count
        )
        result_data += path_results
        metric_data += path_metrics
        print(" done")

    # Concatenate frames
    result_data_frames = [
        pd.DataFrame.from_records(datum, index=[-1]) for datum in result_data
    ]
    result_big_df = pd.concat(result_data_frames, ignore_index=True)

    metric_data_frames = [
        pd.DataFrame.from_records(datum, index=[-1]) for datum in metric_data
    ]
    metric_big_df = pd.concat(metric_data_frames, ignore_index=True)

    # Write to HDF
    result_big_df.to_hdf(output_path, key="main")
    metric_big_df.to_hdf(output_path, key="metrics")


if __name__ == "__main__":
    parser = argparse.ArgumentParser("chia.tools.convert_json_results_to_pandas")
    parser.add_argument("json_path", nargs="+")
    parser.add_argument("--output_path", help="Output file path", required=True)
    parser.add_argument(
        "--accumulate_metric_messages",
        help="Number of accumulated metric messages (default:100)",
        default=100,
    )

    args = parser.parse_args()
    main(
        json_paths=args.json_path,
        output_path=args.output_path,
        max_accumulated_message_count=args.accumulate_metric_messages,
    )



from distutils import util

from setuptools import find_packages, setup

main_ns = {}
ver_path = util.convert_path("chia_tools/version.py")

with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="chia-tools",
    version=main_ns["__version__"],
    packages=["chia_tools"],
    include_package_data=True,
    python_requires=">=3.7",
    install_requires=[
        "tqdm~=4.51",
        "tables~=3.6.1",
        "pandas~=1.0.4",
        "pyqt5~=5.15.0",
        "matplotlib~=3.2.1",
        "chia~=2.0rc20",
    ],
    # metadata to display on PyPI
    author="Clemens-Alexander Brust",
    author_email="clemens-alexander.brust@uni-jena.de",
    description="Extra tools for the CHIA framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    url="https://github.com/cabrust/chia-tools",
)
